/* 
**  mod_openpgp.c -- Apache openpgp module
**  Author: Arturo 'Buanzo' Busleiman <buanzo@buanzo.com.ar>
**  Based on ideas from Apache's mod_authz_host.c, mod_echo.c, mod_access.c
**  Some GPG Code Copyright (C) 2006-2007 Mikael Berthe <mikael@lilotux.net> with Some parts
**  inspired by centericq (impgp.cc)
**
**  Lots of Googling, feedback from modules-dev@apache.org and Kevin Kiley, with a special
**  mention to OWASP, made this possible. Of course, the Apache source also helped :)
**
**  This Apache Module implements a set of OpenPGP-based enhancements to HTTP, such as:
**              * Secure Session Initiation
**              * Request Verification
**              * Session Validation
**
**  Someday it will support encrypted requests/responses, and signed responses. Kevin Kiley's mod_transform.c
**  (included in mod_openpgp's subversion repository) implements the skeleton for encrypted request processing,
**  but is not currently in use.
**
**  Licensed under the Apache License 2.0
*/
/*
use apr_dbm_freedatum()
apr_pstrmemdup(r->pool, dbmval.dptr, dbmval.dsize);
    if (apr_dbm_fetch(dbmfp, dbmkey, &dbmval) == APR_SUCCESS && dbmval.dptr) {
        value = apr_pstrmemdup(r->pool, dbmval.dptr, dbmval.dsize);
    }
    else {
        value = NULL;
    }
dp->locktoken = apr_pmemdup(p, val.dptr + offset, sizeof(*dp->locktoken));
apr_palloc is all over the place. all non-apr allocs have been replaced.... long time ago.
apr_psprintf,
etc
*/
#include <stdio.h>
#include <unistd.h>
#include <httpd.h>
#include <http_core.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_connection.h>
#include <http_log.h>
#include <http_main.h>
#include <http_request.h>
#include <locale.h>
#include <apr_uri.h>
#include <apr_strings.h>
#include <apr_hooks.h>
#include <apr_sha1.h>
#include <apr_md5.h>
#include <apr_dbm.h>
//#include "ap_config.h"

#include <gpgme.h>
#include <gpg-error.h>

//#define DEBUG1
#define MO_NAME "mod_openpgp"
#define MO_VERSION "0.3.0-7.9.0"

//TODO: test ap_is_initial_req
//TODO: use apr_password_get() http://docx.itscales.com/group__apr__lib.html#g377e0677598745769ec6f80fecf8f859

module AP_MODULE_DECLARE_DATA openpgp_module;

/* begin CONFIG section */
typedef struct AuthOpenPGPConfig
{
  int bEnabled;
  char *KeyServer;
  char *HomeDir;
} AuthOpenPGPConfig;

static struct gpg_struct
{
  int enabled;
  char *private_key;
  char *passphrase;
} gpg;

char *SESS_STATUS[] = { "Valid", "Invalid", "Expired" };

#define SESS_STATUS_VALID 0
#define SESS_STATUS_INVALID 1
#define SESS_STATUS_EXPIRED 2

//  gpg_init(priv_key, passphrase)
// Initialize the GPG sub-systems.  This function must be invoked early.
// Note: priv_key & passphrase are optional, they can be set later.
// This function returns 0 if gpgme is available and initialized;
// if not it returns the gpgme error code.
int
gpg_init (const char *priv_key, const char *passphrase)
{
  gpgme_error_t err;

// Check for version and OpenPGP protocol support.
  if (!gpgme_check_version (GPGME_VERSION))
    {
      // ap_log_something  "GPGME initialization error: Bad library version");
      return -1;
    }

  err = gpgme_engine_check_version (GPGME_PROTOCOL_OpenPGP);
  if (err)
    {
      // ap_log_something  "GPGME initialization error: %s", gpgme_strerror (err));
      return err;
    }

// Set the locale information.
  gpgme_set_locale (NULL, LC_CTYPE, setlocale (LC_CTYPE, NULL));
  gpgme_set_locale (NULL, LC_MESSAGES, setlocale (LC_MESSAGES, NULL));

// Store private data.
// if (priv_key!=NULL)
//      gpg_set_private_key (priv_key);
// if (passphrase!=NULL)
//      gpg_set_passphrase (passphrase);

  gpg.enabled = 1;
  return 0;
}

//  strip_header_footer(data)
// Remove PGP header & footer from data.
// Return a new string, or NULL. Uses the request pool. no free() necessary.
static char *
strip_openpgp_header_footer (const char *data, request_rec * r)
{
  char *p, *q;

  if (!data)
    return NULL;

  // p: beginning of real data
  // q: end of real data

  // Strip header (to the first empty line)
  p = strstr (data, "\n\n");
  if (!p)
    return strdup (data);

  // Strip footer
  // We want to remove the last lines, until the line beginning with a '-'
  p += 2;
  for (q = p; *q; q++);
  // (q is at the end of data now)
  for (q--; q > p && (*q != '\n' || *(q + 1) != '-'); q--);

  if (q <= p)
    return NULL;		// Shouldn't happen...

  return apr_pstrndup (r->pool, p, q - p);
//  return strndup (p, q - p);
}


//  passphrase_cb()
// GPGME passphrase callback function.
static gpgme_error_t
passphrase_cb (void *hook, const char *uid_hint, const char *passphrase_info,
	       int prev_was_bad, int fd)
{
  ssize_t len;
  ap_log_cerror (APLOG_MARK, APLOG_NOTICE, 0, NULL,
		 "GPGME: IN_passphrase_cb");
  write (fd, "pepe\n", 5);
  ap_log_cerror (APLOG_MARK, APLOG_NOTICE, 0, NULL,
		 "GPGME: OUT_passphrase_cb");
  return 0;
}

//  (gpg_data)
// Return decrypted gpg_data (or NULL).
// The returned string must be freed with g_free() after use.
char *
gpg_decrypt (const char *gpg_data, conn_rec * c, apr_off_t clength,
	     size_t * nbytes)
{
  gpgme_ctx_t ctx;
  gpgme_data_t in, out;
  char *decrypted_data = NULL;
  size_t nread;
  gpgme_error_t err;
//ap_log_cerror(APLOG_MARK,APLOG_NOTICE,0,c,"GPG_DECRYPT: gpg_data(%d)=%s",clength,gpg_data);
// Check for version and OpenPGP protocol support.
  if (!gpgme_check_version (GPGME_VERSION))
    {
      // ap_log_something  "GPGME initialization error: Bad library version");
      return -1;
    }

  err = gpgme_new (&ctx);
  if (err)
    {
      ap_log_cerror (APLOG_MARK, APLOG_NOTICE, 0, c,
		     "GPG_DECRYPT: gpgme_new: GPGME error: %s",
		     gpgme_strerror (err));
      return NULL;
    }

  gpgme_set_protocol (ctx, GPGME_PROTOCOL_OpenPGP);

/*
	p = getenv("GPG_AGENT_INFO");
	if (!(p && strchr(p, ':')))
	gpgme_set_passphrase_cb(ctx, passphrase_cb, 0);
*/

  gpgme_set_passphrase_cb (ctx, passphrase_cb, 0);

  err = gpgme_data_new_from_mem (&in, gpg_data, clength, 0);
  if (err)
    ap_log_cerror (APLOG_MARK, APLOG_NOTICE, 0, c,
		   "GPG_DECRYPT gpgme_data_new_from_mem: GPGME error: %s",
		   gpgme_strerror (err));
  if (!err)
    {
      err = gpgme_data_new (&out);
      if (err)
	ap_log_cerror (APLOG_MARK, APLOG_NOTICE, 0, c,
		       "GPG_DECRYPT gpgme_data_new: GPGME error: %s",
		       gpgme_strerror (err));
      if (!err)
	{
	  err = gpgme_op_decrypt (ctx, in, out);
	  if (err)
	    ap_log_cerror (APLOG_MARK, APLOG_NOTICE, 0, c,
			   "GPG_DECRYPT gpgme_op_decrypt: GPGME error: %s",
			   gpgme_strerror (err));
	  if (!err)
	    {
	      decrypted_data = gpgme_data_release_and_get_mem (out, &nread);
	      decrypted_data[nread] = 0;
	      if (decrypted_data)
		{
		  *nbytes = nread;
//                      ap_log_cerror(APLOG_MARK,APLOG_NOTICE,0,c,"DECDATA(inside)=%s",decrypted_data);
		  // We need to add a trailing NULL
		  //char *dd = apr_strndup (decrypted_data, nread);
		  //free (decrypted_data);
		  //decrypted_data = dd;
		}
	    }
	  else
	    {
	      gpgme_data_release (out);
	    }
	}
      gpgme_data_release (in);
    }
  if (err && err != GPG_ERR_CANCELED)	// ap_log_err "GPGME decryption error: %s", gpgme_strerror (err));
    if (err)
      ap_log_cerror (APLOG_MARK, APLOG_NOTICE, 0, c,
		     "GPG_DECRYPT err: GPGME error: %s",
		     gpgme_strerror (err));
  gpgme_release (ctx);
  return decrypted_data;
}

//  gpg_encrypt(gpg_data, keyid)
// Return encrypted gpg_data for recipient keyid (or NULL).
// The returned string must be freed with g_free() after use.
char *
gpg_encrypt (const char *gpg_data, const char *keyid, request_rec * r,
	     size_t * nbytes, int ascii, int avoid_traffic_analysis)
{
  gpgme_ctx_t ctx;
  gpgme_data_t in, out;
  char *encrypted_data = NULL, *edata;
  size_t nread;
  gpgme_key_t key, s_key;
  gpgme_error_t err;
// Check for version and OpenPGP protocol support.
  if (!gpgme_check_version (GPGME_VERSION))
    {
      // ap_log_something  "GPGME initialization error: Bad library version");
      return -1;
    }
  err = gpgme_new (&ctx);

  if (err)
    {
      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
		     "GPG_ENCRYPT: gpgme_new: GPGME error: %s",
		     gpgme_strerror (err));
      return NULL;
    }

  gpgme_set_protocol (ctx, GPGME_PROTOCOL_OpenPGP);
  gpgme_set_passphrase_cb (ctx, passphrase_cb, NULL);
  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
		 "GPG_ENCRYPT: set_pass_cb: GPGME error: %s",
		 gpgme_strerror (err));
  gpgme_set_textmode (ctx, 0);
  gpgme_set_armor (ctx, ascii);

  err =
    gpgme_get_key (ctx, "16CCEB319AD39318FF74E40AD9935B0DA494F2DB", &s_key,
		   0);
  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
		 "GPG_ENCRYPT: get_s_key: GPGME error: %s",
		 gpgme_strerror (err));
  err = gpgme_get_key (ctx, keyid, &key, 0);
  if (!err && key)
    {
      gpgme_key_t keys[] = { key, 0 };
      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
		     "GPG_ENCRYPT: Key Found");
      err = gpgme_data_new_from_mem (&in, gpg_data, strlen (gpg_data), 0);
      if (!err)
	{
	  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
			 "GPG_ENCRYPT: loading out data...");
	  err = gpgme_data_new (&out);
	  if (!err)
	    {
	      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
			     "GPG_ENCRYPT: pre signers_add");
	      gpgme_signers_add (ctx, s_key);
	      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
			     "GPG_ENCRYPT: pre op_encrypt_sign");
	      err =
		gpgme_op_encrypt_sign (ctx, keys, GPGME_ENCRYPT_ALWAYS_TRUST,
				       in, out);
	      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
			     "GPG_ENCRYPT: gpgme_op_e_s: GPGME error: %s",
			     gpgme_strerror (err));
	      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
			     "GPG_ENCRYPT: post op_encrypt");
	      if (!err)
		{
		  encrypted_data =
		    gpgme_data_release_and_get_mem (out, &nread);
		  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
				 "GPG_ENCRYPT: 1 - NREAD=%d", nread);
		  if ((encrypted_data) && (nread > 0))
		    encrypted_data[nread] = 0;
		  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
				 "GPG_ENCRYPT: 2");
		  if (encrypted_data)
		    {
		      *nbytes = nread;
		      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
				     "GPG_ENCRYPT: 3");
		    }
		  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
				 "GPG_ENCRYPT: encrypted data is %s",
				 encrypted_data);
		}
	      else
		{
		  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
				 "GPG_ENCRYPT: err at op_encrypt");
		  gpgme_data_release (out);
		}
	    }
	  gpgme_data_release (in);
	}
      gpgme_key_release (key);
    }
  else
    {
/*      scr_LogPrint (LPRINT_LOGNORM, "GPGME encryption error: key not found");*/
      ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
		     "GPG_ENCRYPT: Key NOT found");
      err = 0;
    }
  if (err && err != GPG_ERR_CANCELED)
    ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r, "GPG_ENCRYPT: error: %s",
		   gpgme_strerror (err));
/*    scr_LogPrint (LPRINT_LOGNORM | LPRINT_NOTUTF8,
		  "GPGME encryption error: %s", gpgme_strerror (err));*/
  gpgme_release (ctx);
  ap_log_rerror (APLOG_MARK, APLOG_NOTICE, 0, r,
		 "GPG_ENCRYPT: PRE RETURN encrypted data is %s",
		 encrypted_data);
  return encrypted_data;
}

// gpg_test_passphrase()
// Test the current gpg.passphrase with gpg.private_key.
// If the test doesn't succeed, the passphrase is cleared and a non-null
// value is returned.
int
gpg_test_passphrase (void)
{
/*char *s;

if (!gpg.private_key)
	return -1;	// No private key...

s = gpg_sign ("test");
if (s) {
	free (s);
	return 0;			// Ok, test successful
}
// The passphrase is wrong (if provided)
gpg_set_passphrase (NULL);*/
  return -1;
}

static void *
create_openpgp_server_config (apr_pool_t * p, server_rec * s)
{
  AuthOpenPGPConfig *pConfig = apr_pcalloc (p, sizeof (AuthOpenPGPConfig));

  pConfig->bEnabled = 0;
  pConfig->KeyServer = (char *) apr_pstrdup (p, "hkp://subkeys.pgp.net");
  pConfig->HomeDir = NULL;
  return pConfig;
}

static const char *
openpgp_on (cmd_parms * cmd, void *dummy, int arg)
{
  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (cmd->server->module_config,
			  &openpgp_module);
  pConfig->bEnabled = arg;

  return NULL;
}

static const char *
openpgp_set_keyserver (cmd_parms * cmd, void *dummy, const char *val)
{
  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (cmd->server->module_config,
			  &openpgp_module);
  pConfig->KeyServer = (char *) apr_pstrdup (cmd->pool, val);

  return NULL;
}

static const char *
openpgp_set_homedir (cmd_parms * cmd, void *dummy, const char *val)
{
  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (cmd->server->module_config,
			  &openpgp_module);
  pConfig->HomeDir = (char *) apr_pstrdup (cmd->pool, val);

  return NULL;
}

static const command_rec openpgp_cmds[] = {
  AP_INIT_FLAG ("OpenPGPEngine", openpgp_on, NULL, RSRC_CONF,
		"Enable OpenPGP Auth on this host"),
  AP_INIT_TAKE1 ("OpenPGPKeyServer", openpgp_set_keyserver, NULL, RSRC_CONF,
		 "Specify which keyserver to use for public key imports. Same syntax as gpg\'s --keyserver. Default: hkp://subkeys.pgp.net"),
  AP_INIT_TAKE1 ("OpenPGPHomeDir", openpgp_set_homedir, NULL, RSRC_CONF,
		 "RECOMMENDED - Specify location of .gnupg directory."),
  {NULL}
};

/* end CONFIG section */

gpgme_import_result_t
gpg_import_key (char *key, int msgSize, request_rec * r)
{
  gpgme_ctx_t ctx;
  gpgme_error_t err;
  gpgme_data_t in;
  gpgme_import_result_t result;
  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (r->server->module_config, &openpgp_module);

  if (key == NULL)
    ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r, "NULL KEY");

  gpgme_set_engine_info (GPGME_PROTOCOL_OpenPGP, NULL, pConfig->HomeDir);
  err = gpgme_new (&ctx);
  err = gpgme_data_new_from_mem (&in, key, msgSize, 1);
//ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,key);
  err = gpgme_op_import (ctx, in);
  result = gpgme_op_import_result (ctx);
  gpgme_data_release (in);
  gpgme_release (ctx);
  return (result);
}


int
gpg_verification (char *message, char *signature, int msgSize, char **keyid,
		  char **fpr, char **email, char **name, char **comment,
		  request_rec * r)
{
  gpgme_ctx_t ctx;
  gpgme_error_t err;
  gpgme_data_t sign, text;
  gpgme_verify_result_t result;
  gpgme_key_t key;
  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (r->server->module_config, &openpgp_module);
  int retval = 1;

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"IN gpg_verification");

  if (msgSize < 0)
    return 1;

  gpgme_set_engine_info (GPGME_PROTOCOL_OpenPGP, NULL, pConfig->HomeDir);
  err = gpgme_new (&ctx);

ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): loading text (msgSize = %d, strlen = %d, message= '%s')",msgSize,strlen(message),message);

  if (msgSize == 0) msgSize=strlen(message);
    err = gpgme_data_new_from_mem (&text, message, msgSize, 0);
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): loading signature");

  err = gpgme_data_new_from_mem (&sign, signature, strlen (signature), 0);

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): Verifying...");
  err = gpgme_op_verify (ctx, sign, text, NULL);
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"RESULT = %s, %s",gpgme_strsource (err), gpgme_strerror (err));

// switch(err) {
//      case GPGME_No_Error: {
//              ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): No error");
//              break;
//      }
//      case GPGME_Invalid_Value: {
//              ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): CTX, SIG, PLAIN, or R_STAT is not a valid pointer");
//              break;
//      }
//      case GPGME_No_Data: {
//              ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): SIG does not contain any data to verify");
//              break;
//      }
// }

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): Obtaining Result...");
  result = gpgme_op_verify_result (ctx);
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): Result Obtained (%d) = %s",err,gpgme_strerror(err));

ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): Signature Status (%d) = %s",gpg_err_code (result->signatures->status),gpgme_strerror(result->signatures->status));
apr_table_set(r->headers_in,"X-Auth-OpenPGP-Signature-Status",gpgme_strerror(result->signatures->status));
  if ((result->signatures != NULL) && (result->signatures->fpr != NULL))
    {
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): step1");
      /* copy fingerprint */
      *fpr = apr_pcalloc (r->pool, strlen (result->signatures->fpr) + 1);
      memcpy ((char *) *fpr, result->signatures->fpr,
	      strlen (result->signatures->fpr));
      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): Fingerprint found = %s (LEN=%d)",*fpr,strlen(*fpr));
      //ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): Getting Status");
      switch (gpg_err_code (result->signatures->status))
	{
	case GPG_ERR_NO_ERROR:
	  {
	    err = gpgme_get_key (ctx, result->signatures->fpr, &key, 0);
	    if (key != NULL)
	      {
                              ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"Setting Knownkey -> true");
		apr_table_set (r->headers_in, "X-Auth-OpenPGP-Knownkey","true");
	      }
	    else
	      {
		ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"Setting Knownkey -> false");
		apr_table_set (r->headers_in, "X-Auth-OpenPGP-Knownkey","false");
	      }
	    if (key->subkeys->keyid != NULL)
	      {
		*keyid =
		  apr_pcalloc (r->pool, strlen (key->subkeys->keyid) + 1);
		memcpy ((char *) *keyid, key->subkeys->keyid,
			strlen (key->subkeys->keyid));
	      }
	    if (key->uids->email != NULL)
	      {
		*email = apr_pcalloc (r->pool, strlen (key->uids->email) + 1);
		memcpy ((char *) *email, key->uids->email,
			strlen (key->uids->email));
	      }
	    if (key->uids->name != NULL)
	      {
		*name = apr_pcalloc (r->pool, strlen (key->uids->name) + 1);
		memcpy ((char *) *name, key->uids->name,
			strlen (key->uids->name));
	      }
	    if (key->uids->comment != NULL)
	      {
		*comment =
		  apr_pcalloc (r->pool, strlen (key->uids->comment) + 1);
		memcpy ((char *) *comment, key->uids->comment,
			strlen (key->uids->email));
	      }
	    retval = 0;
	    break;
	  }
	  /*
	     case GPG_ERR_BAD_SIGNATURE:
	     printf("GPGME_SIG_STAT_BAD\n");
	     //*r_stat = GPGME_SIG_STAT_BAD;
	     break;

	     case GPG_ERR_NO_PUBKEY:
	     printf("GPGME_SIG_STAT_NOKEY\n");
	     //*r_stat = GPGME_SIG_STAT_NOKEY;
	     break;

	     case GPG_ERR_NO_DATA:
	     printf("GPGME_SIG_STAT_NOSIG\n");
	     //*r_stat = GPGME_SIG_STAT_NOSIG;
	     break;

	     case GPG_ERR_SIG_EXPIRED:
	     printf("GPGME_SIG_STAT_GOOD_EXP\n");
	     //*r_stat = GPGME_SIG_STAT_GOOD_EXP;
	     break;

	     case GPG_ERR_KEY_EXPIRED:
	     printf("GPGME_SIG_STAT_GOOD_EXPKEY\n");
	     //*r_stat = GPGME_SIG_STAT_GOOD_EXPKEY;
	     break;
	   */
	default:
	  {
	    err = gpgme_get_key (ctx, result->signatures->fpr, &key, 0);
//                      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): get_key_err (%d) = %s",err,gpgme_strerror(err));
	    if (key == NULL)
	      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
			     "RESULT GET KEY is NULL for key");
	    else
	      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
			     "RESULT GET KEY is KNOWN");
	    retval = 1;
	    break;
	  }
	}
    }
  else
    {
      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): NO FINGERPRINT. returning retval=256");
      retval = 256;
    }

// } else {
//      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): result->signatures is NULL.");
//      retval=256;
// }

  gpgme_data_release (sign);
  gpgme_data_release (text);
  gpgme_release (ctx);
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(gpg_verification): retval=%d",retval);
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"OUT gpg_verification");
  return (retval);
}

static int
util_read2 (request_rec * r, char **rbuf)
{
  static int BUFSIZE = 65536;
  unsigned int cantidad = 0;

  if (ap_setup_client_block (r, REQUEST_CHUNKED_DECHUNK) == OK)
    {
      if (ap_should_client_block (r))
	{
	  apr_size_t bytes;
	  char *myBuffer =
	    (char *) apr_pcalloc (r->pool, sizeof (char) * BUFSIZE);
	  do
	    {
	      bytes = ap_get_client_block (r, myBuffer, BUFSIZE);
	      cantidad += bytes;
	    }
	  while (bytes);
	  memcpy ((char *) *rbuf, &myBuffer, cantidad - 1);
	}
    }
}

static int
util_read (request_rec * r, char **rbuf)
{
  int rc = OK;


  if ((rc = ap_setup_client_block (r, REQUEST_CHUNKED_ERROR)) != OK)
    return rc;

  if (ap_should_client_block (r))
    {
      char argsbuffer[HUGE_STRING_LEN];
      int rsize, len_read, rpos = 0;
      long length = r->remaining;

      *rbuf = apr_pcalloc (r->pool, length + 1);

      while ((len_read =
	      ap_get_client_block (r, argsbuffer, sizeof (argsbuffer))) > 0)
	{
	  if ((rpos + len_read) > length)
	    rsize = length - rpos;
	  else
	    rsize = len_read;
	  memcpy ((char *) *rbuf + rpos, argsbuffer, rsize);
	  rpos += rsize;
	}

    }
  return rc;
}


int
remove_session (request_rec * r, const char *sh)
{
//TODO: add remove_session code
}

//char session_status_strings[] = {'Valid','Invalid','Expired'};

int
get_session_status (request_rec * r, char *pgpid, const char *sh)
{				//TODO: use apr_dbd provider and check via sqlite
/*
This function gets the DBM record of key sh, and extracts the PGP KeyID, expire time and remote ip in it.
It then compares those values against the current request values (see below), and determines session status (one of Valid, Invalid or Expired).
MEANING:
	"Good situations": Valid and Expired
		Valid = Remote IP AND pgpid match current request values, and expire_time > now()
		Expired: Remote IP AND pgpid match current request values, but expire_time <= now()
	"ouch situations": INVALUD
		Invalid = Remote IP OR PGPID DO NOT match current request values. Expire time not taken into account for this.

NOTE: "Invalid" is also returned when dbm does not have a record of key = sh.
*/
/* current request values to compare against stored ones */
  return SESS_STATUS_VALID;
  apr_status_t rc;
  apr_dbm_t *db;
  apr_datum_t datum_key = { (char *) sh, (strlen (sh) + 1) };
  apr_datum_t datum_value;
  apr_pool_t *pool;
  int retval = SESS_STATUS_INVALID;	// in case of doubt ;)

  apr_pool_create (&pool, NULL);
  rc =
    apr_dbm_open (&db, "/tmp/mo.sessions.db", APR_DBM_RWCREATE,
		  APR_UREAD | APR_UWRITE, pool);
  if (rc == APR_SUCCESS)
    {
      /*TODO: add checks in here, pseudocode:
         if (key_exists) {
         if ((remote_ip != stored_ip) OR (remote_pgpid != stored_pgpid)) return INVALID;
         if ((remote_ip == stored_ip) AND (remote_pgpid == stored_pgpid)) {
         if (r->request_time >= dbm.timeout_time)
         return EXPIRED;
         else
         return VALID;

         }
         return INVALID; //justincase
         } else {
         return INVALID
         }
       */
      apr_dbm_close (db);
    }
  else
    {
      retval = -1;		//TODO: check for -1 on caller
    }
  apr_pool_destroy (pool);
  return retval;
}

static int
openpgp_end_session_handler (request_rec * r)
{
  const char *xauthopenpgp = NULL;
  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (r->server->module_config, &openpgp_module);

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"IN: end_session_handler %s",r->handler);
  if (!pConfig->bEnabled || !r->handler
      || strcasecmp (r->handler, "mao_EndSession") != 0)
    {
//      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"DECLINE: end_session_handler");
      return DECLINED;
    }

  if (r->method_number != M_GET)
    {
//      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"METH_NOT_ALLOWED: end_session_handler");
      return HTTP_METHOD_NOT_ALLOWED;
    }

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"ACCEPTING: end_session_handler");
  apr_table_set (r->headers_out, "X-OpenPGP-Agent", MO_NAME "/" MO_VERSION);
// Service only for known users. end_session request should be signed by a known key
  xauthopenpgp = apr_table_get (r->headers_in, "X-Auth-OpenPGP");
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(begin_session) X-Auth-OpenPGP=%s",xauthopenpgp);

  apr_table_set (r->headers_out, "X-Auth-OpenPGP", xauthopenpgp);

  if (apr_strnatcmp (xauthopenpgp, "false") == 0)
    {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
		     "(end_session) X-Auth-OpenPGP -> unknown-user (false)");
    }
  else if (apr_strnatcmp (xauthopenpgp, "error") == 0)
    {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
		     "(end_session) Return-Status -> error");
    }
  else
    {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
		     "(end_session) Client KeyID=%s",
		     apr_table_get (r->headers_in, "X-Auth-OpenPGP-KeyID"));
      remove_session (r, apr_table_get (r->headers_in, "X-OpenPGP-Session"));
      ap_set_content_type (r, "text/plain");
      ap_rputs ("END_SESSION_STATUS=destroyed", r);
    }

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"OUT: end_session_handler");
  return OK;
}

static int
import_key_handler (request_rec * r)
{
  char *message = NULL;
  int msgsize = 0;
  gpgme_import_result_t ir;

  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (r->server->module_config, &openpgp_module);
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"IN: import_key_handler");

  if (!pConfig->bEnabled)
    {
      return DECLINED;
    }

  if (!r->handler || strcasecmp (r->handler, "mao_ImportKey") != 0)
    {
      return DECLINED;
    }

  if (r->method_number != M_POST)
    {
      return HTTP_METHOD_NOT_ALLOWED;
    }

  if (util_read (r, &message) != OK)
    return HTTP_METHOD_NOT_ALLOWED;
  if (message == NULL)
    return HTTP_METHOD_NOT_ALLOWED;
  msgsize = atoi (apr_table_get (r->headers_in, "Content-Length"));
  ir = gpg_import_key (message, msgsize, r);

//      fpr = apr_table_get (r->headers_in, "X-Auth-OpenPGP-Fingerprint");
  ap_set_content_type (r, "text/plain");
//ap_rprintf(r,"MessageSize:%d\n",msgsize);
//ap_rprintf(r,"\nPOSTBODY:\n%s|END|\n",message);
  ap_rputs ("HTTP-OpenPGP-ImportStatus:", r);
  if (ir->considered != 1)
    {
      ap_rprintf (r, "Unexpected number of considered keys:%d",
		  ir->considered);
      return OK;
    }
  if (ir->no_user_id != 0)
    {
      ap_rprintf (r, "Unexpected number of user ids:%d", ir->no_user_id);
      return OK;
    }
  if (ir->imported != 0 && ir->imported != 1)
    {
      ap_rputs ("Unexpected number of imported keys", r);
      return OK;
    }
  if (ir->imported_rsa != 0)
    {
      ap_rputs ("Unexpected number of imported RSA keys", r);
      return OK;
    }
  if ((ir->imported == 0 && ir->unchanged != 1)
      || (ir->imported == 1 && ir->unchanged != 0))
    {
      ap_rputs ("Unexpected number of unchanged keys", r);
      return OK;
    }
  if (ir->new_user_ids != 0)
    {
      ap_rputs ("Unexpected number of new user IDs", r);
      return OK;
    }
  if (ir->new_sub_keys != 0)
    {
      ap_rputs ("Unexpected number of new sub keys", r);
      return OK;
    }
  if (ir->new_signatures != 0)
    {
      ap_rputs ("Unexpected number of new signatures", r);
      return OK;
    }
/*if (!ir->imports || ir->imports->next) {
	ap_rputs( "Unexpected number of status reports\n",r);
	return OK;
}*/
  if (ir->imports->result != 0)
    {
      ap_rprintf (r, "Unexpected status result %d\n", ir->imports->result);
      return OK;
    }
  if ((ir->imported == 0 && ir->imports->status != 0)
      || (ir->imported == 1 && ir->imports->status != GPGME_IMPORT_NEW))
    {
      ap_rputs ("Unexpected status", r);
      return OK;
    }
  ap_rputs ("OK", r);
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"OUT: import_key_handler");
  return OK;
}

static int
openpgp_handler (request_rec * r)
{
  const char *openpgp_type = NULL, *openpgp_fields = NULL, *openpgp_sig =
    NULL, *openpgp_algo = NULL, *openpgp_version = NULL, *openpgp_agent =
    NULL, *openpgp_session = NULL;
  char *signature = NULL, *message = NULL, *keyid = NULL, *fingerprint =
    NULL, *email = NULL, *name = NULL, *comment = NULL;
  int msgsize;
  int i = -1;
  int padding = 0;
//apr_uri_t *uri = r->parsed_uri;

  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (r->server->module_config, &openpgp_module);

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,r->unparsed_uri);

  if ((!pConfig->bEnabled)
      || (apr_strnatcmp (r->unparsed_uri, "/HTTP_OPENPGP_IMPORTKEY") == 0))
    {
      return DECLINED;
    }

/* Only accept POST and GET HTTP Methods [for now!] */
  if (r->method_number != M_POST && r->method_number != M_GET)
    return HTTP_METHOD_NOT_ALLOWED;

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"IN: openpgp_handler (URL: %s)",r->unparsed_uri);

/* remove spoofed headers. If module is not enabled for vhost, then if these headers
already exist in the request, they shouldn't be removed, but if it IS enabled... is considered spoofing. */

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): Removing Spoofed X-Auth-OpenPGP Headers (if any)");
  apr_table_unset (r->headers_in, "X-Auth-OpenPGP");
  apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Fingerprint");
  apr_table_unset (r->headers_in, "X-Auth-OpenPGP-KeyID");
  apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Email");
  apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Name");
  apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Comment");
  apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Knownkey");
  apr_table_unset (r->headers_in, "X-OpenPGP-Session-Status"); //TODO: should be X-Auth-OpenPGP-Session-Status

/* read X-OpenPGP headers */
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): Reading Headers");
  openpgp_type = apr_table_get (r->headers_in, "X-OpenPGP-Type");	/* As of Enigform 0.7.9, only S[ign] */
  openpgp_fields = apr_table_get (r->headers_in, "X-OpenPGP-Sig-Fields");	/* as of Enigform 0.7.9, only 'body' */
  openpgp_sig = apr_table_get (r->headers_in, "X-OpenPGP-Sig");
  openpgp_algo = apr_table_get (r->headers_in, "X-OpenPGP-Digest-Algo");	/* not needed with gpgme */
  openpgp_version = apr_table_get (r->headers_in, "X-OpenPGP-Version");
  openpgp_agent = apr_table_get (r->headers_in, "X-OpenPGP-Agent");
  openpgp_session = apr_table_get (r->headers_in, "X-OpenPGP-Session");

  if (openpgp_sig == NULL || openpgp_type == NULL || openpgp_type[0] != 'S')
    {
      apr_table_set (r->headers_in, "X-Auth-OpenPGP", "false");
      return OK;
    }


/* Format Signature */
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): Formatting Signature");
  signature =
    (char *) apr_pstrcat (r->pool, "-----BEGIN PGP SIGNATURE-----\nVersion: ",
			  openpgp_version, "\n\n", openpgp_sig,
			  "\n-----END PGP SIGNATURE------\n", NULL);

/* Append payload to message var, depending on method_number (GET, POST, etc) */

  switch (r->method_number)
    {
    case M_GET:
      {
//              ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): Is GET request");
	message = (char *) apr_pstrcat (r->pool, r->args, NULL);
//	message[strlen(message)]=0;
	msgsize = 0;		// so gpg_verification uses strlen() without -1
	break;
      }
    case M_POST:
      {
//              ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): Is POST request");
	if (util_read (r, &message + padding) != OK)
	  return HTTP_METHOD_NOT_ALLOWED;
	if (message == NULL)
	  return HTTP_METHOD_NOT_ALLOWED;
	if (ap_strstr
	    ((char *) apr_table_get (r->headers_in, "Content-Type"),
	     "boundary="))
	  msgsize =
	    atoi (apr_table_get (r->headers_in, "Content-Length")) - 1;
	else
	  msgsize = atoi (apr_table_get (r->headers_in, "Content-Length"));
	break;
      }
    }

/* Append Session Hash to message var */

//  if ((openpgp_session != NULL) && (ap_strstr(openpgp_fields,"session")))
//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): MALDITA SESION = %s",openpgp_session);
  if (openpgp_session != NULL)
    {
      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): REQUEST HAS SESSION = %s",openpgp_session);
      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): REQUEST HAS MESSAGE = '%s' (%d)",message,strlen(message));
      message = (char *) apr_pstrcat (r->pool, message, openpgp_session,NULL);
      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): MESSAGE+SESSION = '%s' (%d)",message,strlen(message));
      if (msgsize > 0) {
	msgsize += strlen (openpgp_session);
      message[msgsize]=0; }
    }

/* verify the message. */
ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): PRE gpg_verification");
ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): message(strlen=%d): %s",strlen(message),message);
ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): session: %s",openpgp_session);
ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): signature: %s",signature);
ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): msgsize: %d",msgsize);
  i =
    gpg_verification (message, signature, msgsize, &keyid, &fingerprint,
		      &email, &name, &comment, r);
ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): gpg_verification() = %d",i);

//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"(openpgp_handler): Checking fingerprint status");
apr_table_set (r->headers_in, "X-Auth-OpenPGP-Knownkey", "false");

  if (fingerprint != NULL)
    {
      apr_table_set (r->headers_in, "X-Auth-OpenPGP-Fingerprint",
		     fingerprint);
/*      if (strlen (fingerprint) == 16)
	apr_table_set (r->headers_in, "X-Auth-OpenPGP-Knownkey", "false");
      else if (strlen (fingerprint) == 40) */
	apr_table_set (r->headers_in, "X-Auth-OpenPGP-Knownkey", "true");
    }

  if (i == 0)
    {				// GOOD & KNOWN, GOOD & UNKNOWN
      apr_table_set (r->headers_in, "X-Auth-OpenPGP", "true");
      if (keyid != NULL)
	apr_table_set (r->headers_in, "X-Auth-OpenPGP-KeyID", keyid);
      if (email != NULL)
	apr_table_set (r->headers_in, "X-Auth-OpenPGP-Email", email);
      if (name != NULL)
	apr_table_set (r->headers_in, "X-Auth-OpenPGP-Name", name);
      if (comment != NULL)
	apr_table_set (r->headers_in, "X-Auth-OpenPGP-Comment", comment);

      /* Now that we know this request is verified
         we need to check the session hash validity, and append that data
         to the headers */
      if (openpgp_session != NULL)
	{
	  int sess_status = get_session_status (r, keyid, openpgp_session);
//              ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"session_status = %d",sess_status);
	  apr_table_set (r->headers_in, "X-OpenPGP-Session-Status",
			 SESS_STATUS[sess_status]);
	}
      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"OUT: openpgp_handler (true)");
      return OK;
    }
  else
    {
      apr_table_set (r->headers_in, "X-Auth-OpenPGP", "false");
      ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"OUT: openpgp_handler (false)");
      return OK;
    }				/* else { // I consider expired keys, bad signatures as errors. a good, but unknown key see above.
				   apr_table_set(r->headers_in, "X-Auth-OpenPGP", "error");
				   apr_table_set(r->headers_in,"X-Auth-OpenPGP-Knownkey","false");
				   ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"OUT: openpgp_handler (error)");
				   return OK;
				   } */


}				/* end openpgp_handler() */

#ifdef DEBUG
static void
dumpit (ap_filter_t * f, apr_bucket * b)
{
  conn_rec *c = f->c;

  ap_log_error (APLOG_MARK, APLOG_NOTICE, 0, c->base_server,
		"mod_dumpio:  %s (%s-%s): %" APR_SIZE_T_FMT " bytes",
		f->frec->name,
		(APR_BUCKET_IS_METADATA (b)) ? "metadata" : "data",
		b->type->name, b->length);

  if (!(APR_BUCKET_IS_METADATA (b)))
    {
      const char *buf;
      apr_size_t nbytes;
      char *obuf;
      if (apr_bucket_read (b, &buf, &nbytes, APR_BLOCK_READ) == APR_SUCCESS)
	{
	  if (nbytes)
	    {
	      obuf = malloc (nbytes + 1);	/* use pool here. anyway, i'll kill this func I guess */
	      memcpy (obuf, buf, nbytes);
	      obuf[nbytes] = '\0';
	      ap_log_error (APLOG_MARK, APLOG_NOTICE, 0, c->base_server,
			    "mod_dumpio:  %s (%s-%s): %s",
			    f->frec->name,
			    (APR_BUCKET_IS_METADATA (b)) ? "metadata" :
			    "data", b->type->name, obuf);
	      free (obuf);
	    }
	}
      else
	{
	  ap_log_error (APLOG_MARK, APLOG_NOTICE, 0, c->base_server,
			"mod_dumpio:  %s (%s-%s): %s",
			f->frec->name,
			(APR_BUCKET_IS_METADATA (b)) ? "metadata" : "data",
			b->type->name, "error reading data");
	}
    }
}
#endif

#define whichmode( mode ) \
 ( (( mode ) == AP_MODE_READBYTES) ? "readbytes" : \
   (( mode ) == AP_MODE_GETLINE) ? "getline" : \
   (( mode ) == AP_MODE_EATCRLF) ? "eatcrlf" : \
   (( mode ) == AP_MODE_SPECULATIVE) ? "speculative" : \
   (( mode ) == AP_MODE_EXHAUSTIVE) ? "exhaustive" : \
   (( mode ) == AP_MODE_INIT) ? "init" : "unknown" \
 )

static int
openpgp_preconn (conn_rec * c, void *csd)
{
  AuthOpenPGPConfig *pConfig =
    ap_get_module_config (c->base_server->module_config, &openpgp_module);
  return OK;
}

static int
openpgp_init (apr_pool_t * p, apr_pool_t * plog, apr_pool_t * ptemp,
	      server_rec * main_s)
{
  ap_add_version_component (p, MO_NAME "/" MO_VERSION);
  return OK;
}

static void
openpgp_register_hooks (apr_pool_t * p)
{
/* We know SSL uses CONNECTION + 5, so we'll use +4  */
//      ap_register_input_filter("OPENPGP_DECRYPT_INPUT_FILTER", openpgp_decrypt_input_filter, NULL, AP_FTYPE_CONNECTION + 4);
//      ap_hook_pre_connection(openpgp_preconn, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_post_config (openpgp_init, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_post_read_request (openpgp_handler, NULL, NULL, APR_HOOK_FIRST);
//      ap_hook_handler(openpgp_begin_session_handler,NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_handler (import_key_handler, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_handler (openpgp_end_session_handler, NULL, NULL, APR_HOOK_MIDDLE);

  // TODO: check if we need any extra cleanup code for external library gpgme
  //apr_pool_cleanup_register(p, NULL, some_gpgme_terminate_func,apr_pool_cleanup_null);
}

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA openpgp_module = {
  STANDARD20_MODULE_STUFF,
  NULL,				/* create per-dir    config structures */
  NULL,				/* merge  per-dir    config structures */
  create_openpgp_server_config,	/* create_openpgp_server_config create per-server config structures */
  NULL,				/* merge  per-server config structures */
  openpgp_cmds,			/* table of config file commands */
  openpgp_register_hooks	/* register hooks */
};
